use anyhow::Result;
use engine::prelude::*;

use crate::config::experience::ExperienceSettings;

use super::npc_type::NPCType;

pub struct LevelUpEvent;

pub struct Level {
    current_level: u32,
    pub current_experience: u32,
    pub experience_needed: u32,

    exp_settings: ExperienceSettings,
}

impl Clone for Level {
    fn clone(&self) -> Self {
        Self {
            current_level: self.current_level,
            current_experience: self.current_experience,
            experience_needed: self.experience_needed,

            exp_settings: self.exp_settings.clone(),
        }
    }
}

impl Level {
    pub fn new(exp_settings: &ExperienceSettings) -> Self {
        Level {
            current_level: 1,

            current_experience: 0,
            experience_needed: exp_settings.player_experience_needed(1),

            exp_settings: exp_settings.clone(),
        }
    }

    pub fn load(level: u32, current_experience: u32, exp_settings: &ExperienceSettings) -> Self {
        Level {
            current_level: level,

            current_experience,
            experience_needed: exp_settings.player_experience_needed(level),

            exp_settings: exp_settings.clone(),
        }
    }

    /// returns `true` at level up
    pub fn add_experience(&mut self, mob_level: u32, npc_type: impl Into<NPCType>) -> Result<bool> {
        self.add_experience_raw(self.exp_settings.mob_experience(mob_level, npc_type))
    }

    pub fn add_experience_raw(&mut self, exp: u32) -> Result<bool> {
        self.current_experience += exp;

        while self.current_experience >= self.experience_needed {
            self.current_level += 1;

            self.current_experience -= self.experience_needed;
            self.experience_needed = self
                .exp_settings
                .player_experience_needed(self.current_level);

            if self.current_experience < self.experience_needed {
                return Ok(true);
            }
        }

        Ok(false)
    }

    pub fn level(&self) -> u32 {
        self.current_level
    }

    pub fn reset_current(&mut self) {
        self.current_experience = 0;
    }
}

impl EntityComponent for Level {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for Level {
    fn debug_name() -> &'static str {
        "Level"
    }
}
